import logging
import threading
import os
from queue import Queue
from eqsn.shared_dict import SharedDict
from eqsn.qubit_thread import SINGLE_GATE, MERGE_SEND, MERGE_ACCEPT, MEASURE,\
                MEASURE_NON_DESTRUCTIVE, GIVE_STATEVECTOR, \
                CONTROLLED_GATE, NEW_QUBIT, ADD_MERGED_QUBITS_TO_DICT, \
                DOUBLE_GATE, QubitThread


class WorkerProcess(object):

    def __init__(self, queue):
        self.queue = queue
        # Get a new instance, since their might be one from the old process
        self.shared_dict = SharedDict.get_new_instance()

    def run(self):
        """
        Run in loop and wait to receive tasks to perform.
        """
        amount_single_gate = 0
        while True:
            item = self.queue.get()
            if item is None:
                self.stop_all()
                return
            elif item[0] == NEW_QUBIT:
                self.new_qubit(item[1])
            elif item[0] == SINGLE_GATE:
                self.apply_single_gate(item[1], item[2])
                amount_single_gate += 1
            elif item[0] == CONTROLLED_GATE:
                self.apply_controlled_gate(item[1], item[2], item[3])
            elif item[0] == MEASURE:
                self.measure(item[1], item[2])
            elif item[0] == MERGE_ACCEPT:
                self.merge_accept(item[1], item[2])
            elif item[0] == MERGE_SEND:
                self.merge_send(item[1], item[2], item[3])
            elif item[0] == MEASURE_NON_DESTRUCTIVE:
                self.measure_non_destructive(item[1], item[2])
            elif item[0] == ADD_MERGED_QUBITS_TO_DICT:
                self.add_merged_qubits_to_thread(item[1], item[2])
            elif item[0] == GIVE_STATEVECTOR:
                self.give_statevector_for(item[1], item[2])
            elif item[0] == DOUBLE_GATE:
                self.apply_two_qubit_gate(item[1], item[2], item[3])
            else:
                raise ValueError("Command does not exist!")

    def new_qubit(self, q_id):
        """
        Creates a new qubit with an id.

        Args:
            id (String): Id of the new qubit.
        """
        q = Queue()
        thread = QubitThread(q_id, q)
        p = threading.Thread(target=thread.run, args=())
        self.shared_dict.set_thread_with_id(q_id, p, q)
        p.start()
        logging.debug("Created new qubit with id %s.", q_id)

    def measure(self, q_id, channel):
        temp_queue = Queue()
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([MEASURE, q_id, temp_queue])
        res = temp_queue.get()
        channel.put(res)
        self.shared_dict.delete_id_and_check_to_join_thread(q_id)

    def measure_non_destructive(self, q_id, channel):
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([MEASURE_NON_DESTRUCTIVE, q_id, channel])

    def add_merged_qubits_to_thread(self, q_id, qubits):
        q, t = self.shared_dict.get_queues_and_threads_for_ids([q_id])[0]
        for qubit in qubits:
            self.shared_dict.set_thread_with_id(qubit, t, q)

    def stop_all(self):
        """
        Stops the simulator from running.
        """
        self.shared_dict.send_all_threads(None)
        self.shared_dict.stop_all_threads()
        self.shared_dict.stop_shared_dict()

    def apply_two_qubit_gate(self, gate, q_id1, q_id2):
        self.merge_qubits(q_id1, q_id2)
        q = self.shared_dict.get_queues_for_ids([q_id1])[0]
        q.put([DOUBLE_GATE, gate, q_id1, q_id2])

    def apply_single_gate(self, gate, q_id):
        """
        Applies a single gate to a Qubit.
        """
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, gate, q_id])

    def give_statevector_for(self, q_id, channel):
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([GIVE_STATEVECTOR, channel])

    def apply_controlled_gate(self, gate, q_id1, q_id2):
        """
        Applies a controlled gate, where the gate is applied to
        q_id1 and controlled by q_id2.

        Args:
            q_id1 (String): Id of the Qubit on which the X gate is applied.
            q_id2 (String): Id of the Qubit which controls the gate.
        """
        self.merge_qubits(q_id1, q_id2)
        q = self.shared_dict.get_queues_for_ids([q_id1])[0]
        q.put([CONTROLLED_GATE, gate, q_id1, q_id2])

    def merge_send(self, q_id, queue, queue2):
        temp_queue = Queue()
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([MERGE_SEND, queue, temp_queue])
        qubits = temp_queue.get()
        # remove all qubits
        for c in qubits:
            self.shared_dict.delete_id_and_check_to_join_thread(c)
        # send the qubits to the main process
        queue2.put(qubits)

    def merge_accept(self, q_id, queue):
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([MERGE_ACCEPT, queue])

    def merge_qubits(self, q_id1, q_id2):
        """
        Merges two qubits to one process, if they are not already
        running in the same process.

        Args:
            q_id1 (String): Id of the Qubit merged into q_id2.
            q_id2 (String): Id of the Qubit merged with q_id1.
        """
        l = self.shared_dict.get_queues_for_ids([q_id1, q_id2])
        if len(l) == 1:
            return  # Already merged
        else:
            logging.debug("Merge Qubits %s and %s.", q_id1, q_id2)
            q1 = l[0]
            q2 = l[1]
            merge_q = Queue()
            qubits_q = Queue()
            q1.put([MERGE_SEND, merge_q, qubits_q])
            q2.put([MERGE_ACCEPT, merge_q])
            qubits = qubits_q.get()
            self.shared_dict.change_thread_and_queue_of_ids_and_join(
                qubits, q_id2)



import numpy as np
import logging
from copy import deepcopy as dp
import random
import sys

NONE = 0
SINGLE_GATE = 1
CONTROLLED_GATE = 2
MEASURE = 3
MERGE_ACCEPT = 4
MERGE_SEND = 5
MEASURE_NON_DESTRUCTIVE = 7
NEW_QUBIT = 8
ADD_MERGED_QUBITS_TO_DICT = 9
GIVE_STATEVECTOR = 10
DOUBLE_GATE = 11


class QubitThread(object):

    def __init__(self, q_id, queue):
        """
        Args:
            q_id (String): Name of the qubit
            queue (Queue): Queue for receiving commands from main thread.
        """
        # set new seed for random number generator
        local_random = random.Random()
        new_seed = local_random.randrange(1, 100000)
        np.random.seed(new_seed)

        # List of qubits in this thread
        self.qubits = [q_id]

        # receive queue for operations to perform
        self.queue = queue

        # init qubit in state |0>
        self.qubit = np.zeros(2, dtype=np.csingle)
        self.qubit[0] = 1

        logging.debug("Qubit thread with qubit %s has been created.", q_id)

    def apply_single_gate(self, mat, q_id):
        """
        Applys a single gate to a qubit.

        Args:
            mat (np.array): 2x2 unitary array.
            id (String): Qubit on which the gate should be applied to.
        """
        apply_mat = mat
        nr = self.qubits.index(q_id)
        total_amount = len(self.qubits)
        before = nr
        after = total_amount - nr - 1
        if before > 0:
            apply_mat = np.kron(np.eye(2 ** before), apply_mat)
        if after > 0:
            apply_mat = np.kron(apply_mat, np.eye(2 ** after))
        self.qubit = np.dot(apply_mat, self.qubit)

    def give_statevector(self, channel):
        channel.put((dp(self.qubits), dp(self.qubit)))

    def apply_controlled_gate(self, mat, q_id1, q_id2):
        """
        Apply a controlled gate to
        """
        first_mat = 1
        second_mat = 1
        nr1 = self.qubits.index(q_id1)
        nr2 = self.qubits.index(q_id2)
        min_nr = min(nr1, nr2)
        max_nr = max(nr1, nr2)
        if min_nr == nr1:
            # first_mat = mat
            pass
        total_amount = len(self.qubits)
        before = min_nr
        after = total_amount - max_nr - 1
        mid = total_amount - before - after - 2
        if before > 0:
            first_mat = np.eye(2 ** before)
            second_mat = np.eye(2 ** before)
        # Apply first part of Matrix
        if min_nr == nr1:
            first_mat = np.kron(first_mat, np.eye(2))
            second_mat = np.kron(second_mat, mat)
        else:
            first_mat = np.kron(first_mat, np.array([[1, 0], [0, 0]]))
            second_mat = np.kron(second_mat, np.array([[0, 0], [0, 1]]))

        if mid > 0:
            first_mat = np.kron(first_mat, np.eye(2 ** mid))
            second_mat = np.kron(second_mat, np.eye(2 ** mid))
        # Apply second part of Matrix
        if min_nr == nr1:
            first_mat = np.kron(first_mat, np.array([[1, 0], [0, 0]]))
            second_mat = np.kron(second_mat, np.array([[0, 0], [0, 1]]))
        else:
            first_mat = np.kron(first_mat, np.eye(2))
            second_mat = np.kron(second_mat, mat)

        if after > 0:
            first_mat = np.kron(first_mat, np.eye(2 ** after))
            second_mat = np.kron(second_mat, np.eye(2 ** after))
        apply_mat = first_mat + second_mat
        self.qubit = np.dot(apply_mat, self.qubit)

    def merge_accept(self, channel):
        """
        Receive another process to merge it with this one.
        """
        ids = channel.get()
        vector = channel.get()
        self.qubits = self.qubits + ids
        self.qubit = np.kron(self.qubit, vector)
        logging.debug("Qubit Thread merged, new qubits are %r", self.qubits)

    def merge_send(self, channel, chanel2):
        """
        Send own process data to another process and suicide.
        """
        channel.put(dp(self.qubits))
        channel.put(dp(self.qubit))
        chanel2.put(dp(self.qubits))
        return

    def swap_qubits(self, q_id1, q_id2):
        """
        Swaps the position of qubit q_id1 with q_id2
        in the statevector.
        """
        def cnot(q_id1, q_id2):
            mat = np.asarray([[0, 1], [1, 0]])
            self.apply_controlled_gate(mat, q_id1, q_id2)

        # Check if they are the same ids
        if q_id1 == q_id2:
            return

        # Perform swap
        cnot(q_id1, q_id2)
        cnot(q_id2, q_id1)
        cnot(q_id1, q_id2)
        # Change ordering in the list
        i1 = self.qubits.index(q_id1)
        i2 = self.qubits.index(q_id2)
        self.qubits[i1], self.qubits[i2] = self.qubits[i2], self.qubits[i1]

    def apply_two_qubit_gate(self, mat, q_id1, q_id2):
        # Bring the qubits in the right order
        i2 = self.qubits.index(q_id2)
        if i2 > 0:
            new_i1 = i2-1
            self.swap_qubits(q_id1, self.qubits[new_i1])
        else:
            self.swap_qubits(q_id1, self.qubits[0])
            self.swap_qubits(q_id2, self.qubits[1])
        apply_mat = mat
        nr1 = self.qubits.index(q_id1)
        total_amount = len(self.qubits)
        before = nr1
        after = total_amount - nr1 - 2
        if before > 0:
            apply_mat = np.kron(np.eye(2 ** before), apply_mat)
        if after > 0:
            apply_mat = np.kron(apply_mat, np.eye(2 ** after))
        self.qubit = np.dot(apply_mat, self.qubit)

    def measure_non_destructive(self, q_id, ret_channel):
        """
        Perform a non destructive measurement on qubit with the id.
        """
        # determine probability for |1>
        measure_vec = np.array([1, 0], dtype=np.csingle)
        nr = self.qubits.index(q_id)
        total_amount = len(self.qubits)
        before = nr
        after = total_amount - nr - 1
        if before > 0:
            measure_vec = np.kron(np.ones(2 ** before), measure_vec)
        if after > 0:
            measure_vec = np.kron(measure_vec, np.ones(2 ** after))
        pr_0 = np.multiply(measure_vec, self.qubit)
        pr_0 = abs(np.dot(pr_0, pr_0))
        if pr_0 > 1.0:
            pr_0 = 1.0
        elif pr_0 < 0.0:
            pr_0 = 0.0
        meas_res = np.random.binomial(1, 1.0 - pr_0)
        reduction_mat = None
        if meas_res == 0:
            # |0> has been measured
            ret_channel.put(0)
            reduction_mat = np.array([[1, 0], [0, 0]], dtype=np.csingle)
        else:
            # |1> has been measured
            ret_channel.put(1)
            reduction_mat = np.array([[0, 0], [0, 1]], dtype=np.csingle)
        if before > 0:
            reduction_mat = np.kron(
                np.eye(2 ** before, dtype=np.csingle), reduction_mat)
        if after > 0:
            reduction_mat = np.kron(
                reduction_mat, np.eye(2 ** after, dtype=np.csingle))
        # apply measurement result to state vector
        self.qubit = np.dot(reduction_mat, self.qubit)
        # renormalize the qubit vector
        norm = np.linalg.norm(self.qubit)
        self.qubit = self.qubit / norm

    def measure(self, q_id, ret_channel):
        """
        Perform a destructive measurement on qubit with the id.
        """
        # determine probability for |1>
        measure_vec = np.array([1, 0], dtype=np.csingle)
        nr = self.qubits.index(q_id)
        total_amount = len(self.qubits)
        before = nr
        after = total_amount - nr - 1
        if before > 0:
            measure_vec = np.kron(np.ones(2 ** before), measure_vec)
        if after > 0:
            measure_vec = np.kron(measure_vec, np.ones(2 ** after))
        pr_0 = np.multiply(measure_vec, self.qubit)
        pr_0 = abs(np.dot(pr_0, pr_0))
        if pr_0 > 1.0:
            pr_0 = 1.0
        elif pr_0 < 0.0:
            pr_0 = 0.0
        meas_res = np.random.binomial(1, 1.0 - pr_0)
        reduction_mat = None
        if meas_res == 0:
            # |0> has been measured
            ret_channel.put(0)
            reduction_mat = np.array([1, 0], dtype=np.csingle)
        else:
            # |1> has been measured
            ret_channel.put(1)
            reduction_mat = np.array([0, 1], dtype=np.csingle)
        if before > 0:
            reduction_mat = np.kron(
                np.eye(2 ** before, dtype=np.csingle), reduction_mat)
        if after > 0:
            reduction_mat = np.kron(
                reduction_mat, np.eye(2 ** after, dtype=np.csingle))
        self.qubits.remove(q_id)
        if total_amount == 1:
            # it was the last qubit, just terminate this process
            return
        # remove measured qubit from qubit state vector
        self.qubit = np.dot(reduction_mat, self.qubit)
        # renormalize the qubit vector
        norm = np.linalg.norm(self.qubit)
        self.qubit = self.qubit / norm

    def run(self):
        """
        Run in loop and wait to receive tasks to perform.
        """
        while True:
            item = self.queue.get()
            if item is None:
                return
            elif item[0] == SINGLE_GATE:
                self.apply_single_gate(item[1], item[2])
            elif item[0] == CONTROLLED_GATE:
                self.apply_controlled_gate(item[1], item[2], item[3])
            elif item[0] == MEASURE:
                self.measure(item[1], item[2])
                # no qubit left, terminate
                if len(self.qubits) == 0:
                    return
            elif item[0] == MERGE_ACCEPT:
                self.merge_accept(item[1])
            elif item[0] == MERGE_SEND:
                self.merge_send(item[1], item[2])
                return
            elif item[0] == MEASURE_NON_DESTRUCTIVE:
                self.measure_non_destructive(item[1], item[2])
            elif item[0] == GIVE_STATEVECTOR:
                self.give_statevector(item[1])
            elif item[0] == DOUBLE_GATE:
                self.apply_two_qubit_gate(item[1], item[2], item[3])
            else:
                raise ValueError("Command does not exist!")



import multiprocessing
import logging
import numpy as np
from eqsn.qubit_thread import SINGLE_GATE, MERGE_SEND, MERGE_ACCEPT, MEASURE,\
                MEASURE_NON_DESTRUCTIVE, GIVE_STATEVECTOR, DOUBLE_GATE, \
                CONTROLLED_GATE, NEW_QUBIT, ADD_MERGED_QUBITS_TO_DICT
from eqsn.shared_dict import SharedDict
from eqsn.worker_process import WorkerProcess
from eqsn.process_picker import ProcessPicker


class EQSN(object):
    """
    Main object of EQSN, with this object, all of the Qubits can be controlled.
    All functions are threadsafe, but at the moment, only one instance should be
    used.
    """
    __instance = None

    @staticmethod
    def get_instance():
        if EQSN.__instance is None:
            return EQSN()
        return EQSN.__instance

    def __init__(self):
        if EQSN.__instance is not None:
            raise ValueError("Use get instance to get this class")
        EQSN.__instance = self
        self.manager = multiprocessing.Manager()
        self.shared_dict = SharedDict.get_instance()
        cpu_count = multiprocessing.cpu_count()
        self.process_queue_list = []
        for _ in range(cpu_count):
            q = multiprocessing.Queue()
            new_worker = WorkerProcess(q)
            p = multiprocessing.Process(target=new_worker.run, args=())
            p.start()
            self.process_queue_list.append((p, q))
        self.process_picker = ProcessPicker.get_instance(
            cpu_count, self.process_queue_list)

    def new_qubit(self, q_id):
        """
        Creates a new qubit with an id.

        Args:
            q_id (String): Id of the new qubit.
        """
        p, q = self.process_picker.get_next_process_queue()
        q.put([NEW_QUBIT, q_id])
        self.shared_dict.set_thread_with_id(q_id, p, q)
        logging.debug("Created new qubit with id %s.", q_id)

    def stop_all(self):
        """
        Stops the simulator from running.
        """
        for p, q in self.process_queue_list:
            q.put(None)
            p.join()
        self.shared_dict.stop_shared_dict()
        self.process_picker.stop_process_picker()
        EQSN.__instance = None

    def X_gate(self, q_id):
        """
        Applies the Pauli X gate to the Qubit with q_id.
        """
        x = np.array([[0, 1], [1, 0]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def Y_gate(self, q_id):
        """
        Applies the Pauli Y gate to the Qubit with q_id.
        """
        x = np.array([[0, 0 - 1j], [0 + 1j, 0]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def Z_gate(self, q_id):
        """
        Applies the Pauli Z gate to the Qubit with q_id.
        """
        x = np.array([[1, 0], [0, -1]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def H_gate(self, q_id):
        """
        Applies the Hadamard gate to the Qubit with q_id.
        """
        x = (1 / 2.0) ** 0.5 * np.array([[1, 1], [1, -1]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def T_gate(self, q_id):
        """
        Applies the T gate to the Qubit with q_id.
        """
        x = np.array(
            [[1, 0], [0, (0.7071067811865476 + 0.7071067811865475j)]],
            dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def S_gate(self, q_id):
        """
        Applies the S gate to the Qubit with q_id.
        """
        x = np.array([[1, 0], [0, 1j]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def K_gate(self, q_id):
        """
        Applies the K gate to the Qubit with q_id.
        """
        x = 0.5 * np.array([[1+1j, 1-1j], [-1+1j, -1-1j]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def RX_gate(self, q_id, rad):
        """
        Applies the T gate to the Qubit with q_id.
        """
        mid = np.cos(rad / 2)
        other = -1j * np.sin(rad / 2)
        x = np.array([[mid, other], [other, mid]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def RY_gate(self, q_id, rad):
        """
        Applies the T gate to the Qubit with q_id.
        """
        mid = np.cos(rad / 2)
        other = np.sin(rad / 2)
        x = np.array([[mid, -1.0 * other], [other, mid]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def RZ_gate(self, q_id, rad):
        """
        Applies the T gate to the Qubit with q_id.
        """
        top = np.exp(-1j * (rad / 2))
        bot = np.exp(1j * (rad / 2))
        x = np.array([[top, 0], [0, bot]], dtype=np.csingle)
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, x, q_id])

    def custom_gate(self, q_id, gate):
        """
        Applies a custom gate to the qubit with q_id.
        """
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([SINGLE_GATE, gate, q_id])

    def merge_qubits(self, q_id1, q_id2):
        """
        Merges two qubits to one process, if they are not already
        running in the same process.

        Args:
            q_id1 (String): Id of the Qubit merged into q_id2.
            q_id2 (String): Id of the Qubit merged with q_id1.
        """
        l = self.shared_dict.get_queues_for_ids([q_id1, q_id2])
        if len(l) == 1:
            return  # Already merged
        else:
            # Block the dictionary, that noone can send commands to the qubits,
            logging.debug("Merge Qubits %s and %s.", q_id1, q_id2)
            self.shared_dict.block_shared_dict()
            q1 = l[0]
            q2 = l[1]
            merge_q = self.manager.Queue()
            qubits_q = self.manager.Queue()
            q1.put([MERGE_SEND, q_id1, merge_q, qubits_q])
            q2.put([MERGE_ACCEPT, q_id2, merge_q])
            qubits = qubits_q.get()
            q2.put([ADD_MERGED_QUBITS_TO_DICT, q_id2, qubits])
            self.shared_dict.change_thread_and_queue_of_ids_nonblocking(
                qubits, q_id2)
            self.shared_dict.release_shared_dict()

    def cnot_gate(self, q_id1, q_id2):
        """
        Applies a controlled X gate, where the gate is applied to
        q_id1 and controlled by q_id2.

        Args:
            q_id1 (String): Id of the Qubit on which the X gate is applied.
            q_id2 (String): Id of the Qubit which controls the gate.
        """
        x = np.array([[0, 1], [1, 0]], dtype=np.csingle)
        self.merge_qubits(q_id1, q_id2)
        q = self.shared_dict.get_queues_for_ids([q_id1])[0]
        q.put([CONTROLLED_GATE, x, q_id1, q_id2])

    def cphase_gate(self, q_id1, q_id2):
        """
        Applies a controlled Z gate, where the gate is applied to
        q_id1 and controlled by q_id2.
        """
        x = np.array([[0, 1], [0, -1]], dtype=np.csingle)
        self.merge_qubits(q_id1, q_id2)
        q = self.shared_dict.get_queues_for_ids([q_id1])[0]
        q.put([CONTROLLED_GATE, x, q_id1, q_id2])

    def give_statevector_for(self, q_id):
        ret = self.manager.Queue()
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        q.put([GIVE_STATEVECTOR, q_id, ret])
        qubits, vector = ret.get()
        return qubits, vector

    def custom_two_qubit_gate(self, q_id1, q_id2, gate):
        self.merge_qubits(q_id1, q_id2)
        q = self.shared_dict.get_queues_for_ids([q_id1])[0]
        q.put([DOUBLE_GATE, gate, q_id1, q_id2])

    def custom_controlled_gate(self, applied_to_id, controlled_by_id, gate):
        self.merge_qubits(applied_to_id, controlled_by_id)
        q = self.shared_dict.get_queues_for_ids([applied_to_id])[0]
        q.put([CONTROLLED_GATE, gate, applied_to_id, controlled_by_id])

    def measure(self, q_id, non_destructive=False):
        """
        Measures a qubit with an id. If non_destructive is False, the qubit
        is removed from the system, otherwise, the qubit stays in the system
        after measurement, but its wavefunction collapses.

        Args:
            id (String): Id of the Qubit which should be measured.
            non_destructive(bool): If a qubit should not be removed from the
                                    system after measurement.
        """
        ret = self.manager.Queue()
        q = self.shared_dict.get_queues_for_ids([q_id])[0]
        if non_destructive:
            q.put([MEASURE_NON_DESTRUCTIVE, q_id, ret])
        else:
            q.put([MEASURE, q_id, ret])
        res = ret.get()
        if not non_destructive:
            self.shared_dict.delete_id_and_check_to_join_thread(q_id)
        logging.debug(
            "Qubit with id %s has been measured with outcome %d.", q_id, res)
        return res



class ProcessPicker(object):
    """
    Decides which process allocates new Qubits.
    """
    __instance = None

    @staticmethod
    def get_instance(cpu_count, process_queue_list):
        if ProcessPicker.__instance is None:
            return ProcessPicker(cpu_count, process_queue_list)
        return ProcessPicker.__instance

    def __init__(self, cpu_count, process_queue_list):
        if ProcessPicker.__instance is not None:
            raise ValueError(
                "Use get instance to get the Process picker class.")
        ProcessPicker.__instance = self
        self.amount_processes = cpu_count
        self.pointer = 0
        self.process_queue_list = process_queue_list

    def get_next_process_queue(self):
        """
        Called in threads may override pointer, but we do not care, since
        the pointer value is not essential.
        """
        res_q = self.process_queue_list[self.pointer % self.amount_processes]
        self.pointer += 1
        return res_q

    def stop_process_picker(self):
        ProcessPicker.__instance = None



import threading


class SharedDict(object):

    # From O'Reilly Python Cookbook by David Ascher, Alex Martelli
    # With changes to cover the starvation situation where a continuous
    #   stream of readers may starve a writer, Lock Promotion and Context Managers
    class ReadWriteLock:
        """
        A lock object that allows many simultaneous "read locks", but
        only one "write lock."
        """

        def __init__(self, with_promotion=False):
            self._read_ready = threading.Condition(threading.RLock())
            self._readers = 0
            self._writers = 0
            self._promote = with_promotion
            self._readerList = []  # List of Reader thread IDs
            self._writerList = []  # List of Writer thread IDs

        def acquire_read(self):
            """
            Acquire a read lock. Blocks only if a thread has
            acquired the write lock.
            """
            self._read_ready.acquire()
            try:
                while self._writers > 0:
                    self._read_ready.wait()
                    self._readers += 1
            finally:
                self._readerList.append(threading.get_ident())
                self._read_ready.release()

        def release_read(self):
            """
            Release a read lock.
            """
            self._read_ready.acquire()
            try:
                self._readers -= 1
                if not self._readers:
                    self._read_ready.notifyAll()
            finally:
                self._readerList.remove(threading.get_ident())
                self._read_ready.release()

        def acquire_write(self):
            """
            Acquire a write lock. Blocks until there are no
            acquired read or write locks.
            """
            self._read_ready.acquire()
            self._writers += 1
            self._writerList.append(threading.get_ident())
            while self._readers > 0:
                if self._promote and threading.get_ident() in self._readerList and set(self._readerList).issubset(
                        set(self._writerList)):
                    break
                else:
                    self._read_ready.wait()

        def release_write(self):
            """
            Release a write lock.
            """
            self._writers -= 1
            self._writerList.remove(threading.get_ident())
            self._read_ready.notifyAll()
            self._read_ready.release()

    __instance = None

    @staticmethod
    def get_instance():
        if SharedDict.__instance is None:
            return SharedDict()
        return SharedDict.__instance

    @staticmethod
    def get_new_instance():
        SharedDict.__instance = None
        return SharedDict()

    def __init__(self):
        if SharedDict.__instance is not None:
            raise Exception("Call get instance to get this class!")
        SharedDict.__instance = self
        self.lock = SharedDict.ReadWriteLock()
        self.id_to_queue = {}

        self.id_to_thread = {}
        self.thread_list = []
        self.queue_list = []

    def block_shared_dict(self):
        self.lock.acquire_write()

    def release_shared_dict(self):
        self.lock.release_write()

    def get_queues_and_threads_for_ids(self, q_id_list):
        ret = []
        self.lock.acquire_read()
        for q_id in q_id_list:
            res = self.id_to_queue[q_id]
            res2 = self.id_to_thread[q_id]
            if res not in ret:
                ret.append((res, res2))
        self.lock.release_read()
        return ret

    def get_queues_for_ids(self, q_id_list):
        ret = []
        self.lock.acquire_read()
        for q_id in q_id_list:
            res = self.id_to_queue[q_id]
            if res not in ret:
                ret.append(res)
        self.lock.release_read()
        return ret

    def set_thread_with_id(self, q_id, thread, queue):
        self.lock.acquire_write()
        self.id_to_queue[q_id] = queue
        self.id_to_thread[q_id] = thread
        self.queue_list.append(queue)
        self.thread_list.append(thread)
        self.lock.release_write()

    def delete_id(self, q_id):
        self.lock.acquire_write()
        del self.id_to_queue[q_id]
        del self.id_to_thread[q_id]
        self.lock.release_write()

    def delete_id_and_check_to_join_thread(self, q_id):
        self.lock.acquire_write()
        thread = None
        if q_id in self.id_to_thread.keys():
            thread = self.id_to_thread[q_id]
            if not thread.is_alive():
                thread.join()
                if thread in self.thread_list:
                    self.thread_list.remove(thread)
            del self.id_to_thread[q_id]
            del self.id_to_queue[q_id]
        self.lock.release_write()

    def join_thread_with_id(self, q_id):
        self.lock.acquire_read()
        thread = self.id_to_thread[q_id]
        self.lock.release_read()
        thread.join()
        self.lock.acquire_write()
        self.thread_list.remove(thread)
        del self.id_to_thread[q_id]
        self.lock.release_write()

    def change_thread_and_queue_of_ids_and_join(self, q_ids, q_id_new_thread):
        self.lock.acquire_write()
        new_thread = self.id_to_thread[q_id_new_thread]
        new_queue = self.id_to_queue[q_id_new_thread]
        for q_id in q_ids:
            thread = self.id_to_thread[q_id]
            if thread in self.thread_list:
                self.thread_list.remove(thread)
                thread.join()
            self.id_to_thread[q_id] = new_thread
            self.id_to_queue[q_id] = new_queue
        self.lock.release_write()

    def change_thread_and_queue_of_ids(self, q_ids, q_id_new_thread):
        self.lock.acquire_write()
        new_thread = self.id_to_thread[q_id_new_thread]
        new_queue = self.id_to_queue[q_id_new_thread]
        for q_id in q_ids:
            self.id_to_thread[q_id] = new_thread
            self.id_to_queue[q_id] = new_queue
        self.lock.release_write()

    def change_thread_and_queue_of_ids_nonblocking(self, q_ids, q_id_new_thread):
        new_thread = self.id_to_thread[q_id_new_thread]
        new_queue = self.id_to_queue[q_id_new_thread]
        for q_id in q_ids:
            self.id_to_thread[q_id] = new_thread
            self.id_to_queue[q_id] = new_queue

    def send_all_threads(self, msg):
        self.lock.acquire_write()
        for p in self.queue_list:
            p.put(msg)
        self.lock.release_write()

    def stop_all_threads(self):
        self.lock.acquire_write()
        for p in self.thread_list:
            p.join()
        self.lock.release_write()

    def stop_shared_dict(self):
        SharedDict.__instance = None



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='EQSN',
    version='0.0.6',
    packages=['eqsn', ],
    license='MIT',
    author='Benjamin Zanger',
    author_email='benjamin.zanger@tum.de',
    description="A quantum simulator made for networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/tqsd/EQSN_python',
    download_url='https://github.com/tqsd/EQSN_python/releases/tag/v0.0.6',
    keywords=['quantum-simulator', 'quantum-internet', 'QuNetSim'],
    install_requires=[
        'numpy',
      ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    python_requires='>=3.6',
)
